package net.sf.sidaof.dao.jpa;

import java.util.List;

import javax.persistence.EntityNotFoundException;
import javax.persistence.LockModeType;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.PersistenceException;
import javax.persistence.Query;

import net.sf.sidaof.dao.AbstractDao;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Base class for JPA DAOs. To use, extend it in DAO classes and implement
 * {@link #getEntityManager()}.
 * 
 * {@link #getEntityManager()} is externalized to allow subclasses to obtain the
 * correct entity manager, particularly when using multiple persistence units,
 * and in the necessary manner, e.g. specifying the unitName:
 * 
 * <pre>
 * &#064;PersistenceContext(unitName = &quot;persistenceUnitName&quot;)
 * private EntityManager entityManager;
 * </pre>
 * 
 * @param <K>
 *            The primary key class of the entity, e.g. Long.
 * @param <E>
 *            The entity class the DAO persists, e.g. Person.
 * 
 * @see net.sf.sidaof.dao.jpa.JpaDaoImpl
 */
public abstract class AbstractJpaDao<K, E> extends AbstractDao<K, E> implements
        JpaDao<K, E> {
    private static final Logger LOG =
            LoggerFactory.getLogger(AbstractJpaDao.class);

    /** Make a new instance, automatically determining the entity class. */
    public AbstractJpaDao() {
        super();
    }

    /**
     * Make a new instance, specifying the entity class.
     * 
     * @param entityClass
     *            The entity class to manage.
     */
    public AbstractJpaDao(Class<E> entityClass) {
        super(entityClass);
    }

    @Override
    public E findById(K id) {
        LOG.debug("findById: id={}", id);
        return getEntityManager().find(entityClass, id);
    }

    @Override
    public E findById(K id, LockModeType lockMode, boolean exceptionIfNotFound) {
        final Object[] logArgs = {id, lockMode, exceptionIfNotFound};
        LOG.debug("findById: id={}, lock={}, exceptionIfNotFound={}", logArgs);
        E entity = findById(id);
        if (entity == null) {
            final String msg =
                    "Entity " + entityClass.getName() + " with id " + id
                            + " not found";
            LOG.info("getById: {}", msg);
            if (exceptionIfNotFound) {
                throw new EntityNotFoundException(msg);
            }
        } else if (lockMode != null) {
            // don't try lock if entity is null
            getEntityManager().lock(entity, lockMode);
        }
        return entity;
    }

    @Override
    @SuppressWarnings("unchecked")
    public List<E> findAll() {
        Query query =
                getEntityManager().createQuery("from " + entityClass.getName());
        LOG.debug("findAll: query={}", query);
        List<E> list = query.getResultList();
        return list;
    }

    @Override
    @SuppressWarnings("unchecked")
    public List<E> findAll(int startPosition, int maxResult) {
        LOG.debug("findAll: startPosition={}, maxResult={}", startPosition,
                maxResult);
        Query query =
                getEntityManager().createQuery("from " + entityClass.getName())
                        .setFirstResult(startPosition).setMaxResults(maxResult);
        LOG.debug("findAll: query={}", query);
        List<E> list = query.getResultList();
        return list;
    }

    @Override
    public void persist(E entity) {
        LOG.debug("persist: entity={}", entity);
        getEntityManager().persist(entity);
        LOG.debug("persist: persisted entity={}", entity);
    }

    @Override
    public E merge(E entity) {
        LOG.debug("merge: entity={}", entity);
        E newEntity = getEntityManager().merge(entity);
        LOG.debug("merge: merged entity={}", newEntity);
        return newEntity;
    }

    @Override
    public void remove(E entity) {
        LOG.debug("remove: entity={}", entity);
        getEntityManager().remove(entity);
    }

    @Override
    public void clear() {
        LOG.debug("clear: called");
        getEntityManager().clear();
    }

    @Override
    public void flush() {
        LOG.debug("flush: called");
        getEntityManager().flush();
    }

    @Override
    public boolean contains(E entity) {
        LOG.debug("contains: entity={}", entity);
        boolean contains = getEntityManager().contains(entity);
        return contains;
    }

    @Override
    public void refresh(E entity) {
        LOG.debug("refresh: old entity={}", entity);
        getEntityManager().refresh(entity);
        LOG.debug("refresh: new entity={}", entity);
    }

    @Override
    public Long countAll(String where, String pathExpression) {
        if (where == null) {
            where = "";
        } else {
            where = " where " + where;
        }
        if (pathExpression == null) {
            pathExpression = "*";
        }

        final StringBuilder sb = new StringBuilder(1000);
        sb.append("select count(").append(pathExpression).append(") from ");
        sb.append(getEntityClass().getName());
        sb.append(" e");
        sb.append(where);

        String queryString = sb.toString();
        LOG.debug("countAll: queryString={}", queryString);

        Query q = getEntityManager().createQuery(queryString);
        Long count = (Long) q.getSingleResult();
        return count;
    }

    @Override
    public E runQueryWithSingleResult(Query q) {
        LOG.debug("runQueryWithSingleResult: query={}", q);
        return runQueryWithSingleResult(q, null, false);
    }

    @Override
    @SuppressWarnings("unchecked")
    public E runQueryWithSingleResult(Query q, LockModeType lockMode,
            boolean exceptionIfNotFound) {
        final Object[] logArgs = {q, lockMode, exceptionIfNotFound};
        LOG.debug("runQueryWithSingleResult:"
                + " query={}, lock={}, exceptionIfNotFound={}", logArgs);

        E entity = null;
        try {
            entity = (E) q.getSingleResult();
            if (lockMode != null) {
                getEntityManager().lock(entity, lockMode);
            }
        } catch (NoResultException e) {
            handleQueryException(QUERY_SINGLERESULT_NOTFOUND_LOGMSG, e, q,
                    exceptionIfNotFound);
        } catch (NonUniqueResultException e) {
            handleQueryException(QUERY_SINGLERESULT_MULTFOUND_LOGMSG, e, q,
                    exceptionIfNotFound);
        }
        return entity;
    }

    private void handleQueryException(String msg, PersistenceException e,
            Query q, boolean exceptionIfNotFound) {
        Object[] logArgs = {entityClass.getName(), q};
        LOG.warn(msg, logArgs, e);
        if (exceptionIfNotFound) {
            throw e;
        }
    }
}
